"""Exceptions used by pywam."""
from __future__ import annotations


class PywamError(Exception):
    """ General pywam exception. """


class FeatureNotSupportedError(PywamError):
    """ Feature not supported by current mode or app. """


class ApiCallError(PywamError):
    """ Api call was not successful. """


class ApiCallTimeoutError(PywamError):
    """ Api call response time out. """



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Samsung Wireless Audio speaker (WAM). """
from __future__ import annotations

import asyncio
import functools
import logging
from typing import TYPE_CHECKING

from pywam.client import WamClient
from pywam.device import WamDevice, get_device_info
from pywam.events import WamEvents
from pywam.attributes import WamAttributes
from pywam.lib import api_call, validate
from pywam.lib.const import (
    APP_FEATURES,
    EXC_MESSAGE,
    Feature,
    SOURCES_BY_API,
    SOURCES_BY_NAME,
    SOURCE_FEATURES,
)
from pywam.lib.equalizer import EqualizerPreset
from pywam.lib.exceptions import ApiCallError, FeatureNotSupportedError, PywamError


if TYPE_CHECKING:
    from pywam.lib.api_response import ApiResponse
    from pywam.lib.media_presets import MediaPreset
    from pywam.lib.url import UrlMediaItem


_LOGGER = logging.getLogger(__name__)


def is_it_supported(func):
    """ Check if feature is supported in current state.

    We should always make sure that speaker support's the things
    we are to send to minimize the risk of "bricking" the device.

    Raises:
        FeatureNotSupportedError: If feature is not supported.
    """

    @functools.wraps(func)
    def wrapper_is_it_supported(self, *args, **kwargs):
        if func.__name__ not in self.supported_features:
            raise FeatureNotSupportedError(
                f'{EXC_MESSAGE[func.__name__]} is not supported in this mode.'
            )
        return func(self, *args, **kwargs)
    return wrapper_is_it_supported


class Speaker():
    """ Represents a Samsung Wireless Audio speaker (Samsung WAM).

    Arguments:
        ip:
            Speaker's IP address. (e.g "192.168.1.100")
        port (optional):
            TCP port that speaker listens on.
        user (optional):
            UUID for the connected user.
    """

    def __init__(self,
                 ip: str,
                 port: int = 55001,
                 user: str = '',
                 ) -> None:
        """ Initialize new speaker object. """
        self._ip: str = validate.ip(ip)
        self._port: int = validate.port(port)
        self._user: str = validate.user(user)
        self.device = WamDevice()
        self.attribute = WamAttributes(self, self.device)
        self.events = WamEvents(self)
        self.client = WamClient(self)

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb) -> None:
        await self.disconnect()

    @property
    def ip(self) -> str:
        """ Return speaker ip address. """
        return self._ip

    @property
    def port(self) -> int:
        """ Return which TCP port speaker listens on. """
        return self._port

    @property
    def user(self) -> str:
        """ Return user id (UUID). """
        return self._user

    @property
    def supported_features(self) -> list[str]:
        """ Return a list of current supported features. """

        supported_features: list[str] = []

        # If speaker is slave in group only basic features are supported.
        if self.attribute.is_slave:
            return supported_features

        # Pause is supported by all services.
        # Select source is supported unless speaker is slave
        supported_features.extend([Feature.PAUSE, Feature.SELECT_SOURCE])

        # If we don't know all the needed attributes, we should return now.
        if self.attribute._grouptype is None or self.attribute._function is None:
            return supported_features

        # App and source dependant features
        if self.attribute.app_name:
            supported_features.extend(APP_FEATURES.get(self.attribute.app_name, []))
        if self.attribute.source:
            supported_features.extend(SOURCE_FEATURES.get(self.attribute.source, []))

        # If speaker is not grouped (master or slave) everything is supported
        if not self.attribute.is_master:
            supported_features.extend([Feature.SET_NAME])

        return supported_features

    # ******************************************************************
    # Connection
    # ******************************************************************
    async def connect(self,
                      connection_timeout: int = 5,
                      request_timeout: int = 5,
                      ) -> None:
        """ Connect to speaker.

        Arguments:
            connection_timeout (optional):
                Timeout in seconds for speaker connection.
            request_timeout (optional):
                Timeout when sending API commands to the speaker.

        Raises:
            ConnectionError: If no initial connection can be established.
        """
        self.client.request_timeout = request_timeout
        self.client.connection_timeout = connection_timeout
        self.client.register_subscriber(self.events.receiver)
        await self.client.connect()
        await self.client.start_listening()

    async def disconnect(self) -> None:
        """ Disconnect from speaker. """
        await self.client.stop_listening()
        await self.client.disconnect()
        self.client.unregister_subscriber(self.events.receiver)

    @property
    def is_connected(self) -> bool:
        """ Return True if connected to speaker, otherwise False. """
        if self.client.is_connected and self.client.is_listening:
            return True
        return False

    # ******************************************************************
    # Commands
    # ******************************************************************
    @is_it_supported
    async def cmd_next(self) -> None:
        """ Send next track command to speaker."""
        await self.client.request(api_call.set_trick_mode('next'))

    @is_it_supported
    async def cmd_pause(self) -> None:
        """ Send pause command to speaker. """
        if self.attribute._submode == 'dlna':
            await self.client.request(api_call.set_uic_playback_control('pause'))
        elif self.attribute._submode == 'cp':
            await self.client.request(api_call.set_cpm_playback_control('pause'))
        if self.attribute._submode == 'url' or self.attribute.app_name == 'Unknown':
            # For url_playback 'pause' is supported but not 'stop'.
            # But I have not found any way to resume it.
            # 'resume', 'play' and 'stop' is not supported for url_playback!
            # It has happened that it has resumed by it self hours after
            # it was paused. So until we find out how to resume it we
            # pause the stream and mute the speaker, and then set state
            # to stop.
            pause = await self.client.request(api_call.set_uic_playback_control('pause'))
            if pause.success:
                mute = await self.client.request(api_call.set_mute(True))
                if mute.success:
                    self.attribute._playstatus = 'stop'
                    self.attribute._cpname = 'Unknown'
                    self.attribute._submode = 'cp'

    @is_it_supported
    async def cmd_play(self) -> None:
        """ Send play command to speaker. """
        if self.attribute._submode == 'cp':
            await self.client.request(api_call.set_cpm_playback_control('play'))
        elif self.attribute._submode == 'dlna':
            await self.client.request(api_call.set_uic_playback_control('resume'))

    @is_it_supported
    async def cmd_previous(self) -> None:
        """ Send previous track command to speaker."""
        await self.client.request(api_call.set_trick_mode('previous'))

    @is_it_supported
    async def cmd_stop(self) -> None:
        """ Send stop command to speaker. """
        await self.client.request(api_call.set_cpm_playback_control('stop'))

    # ******************************************************************
    # Play
    # ******************************************************************
    @is_it_supported
    async def play_preset(self, preset: MediaPreset) -> None:
        """ Play TuneIn preset stored in speaker.

        Arguments:
            preset:
                MediaPreset object to play. Get available presets
                with :attr:`Speaker.attribute.media_presets`
        """
        try:
            preset = validate.media_preset(preset, self.attribute.tunein_presets)
        except ValueError as e:
            # If preset is not found on the speaker we update presets
            await self.client.request(api_call.set_select_radio())
            await self.client.request(api_call.get_preset_list(0, 30))
            raise e

        # At the moment only TuneIn presets are supported.
        if preset.app != 'TuneIn':
            raise FeatureNotSupportedError('Only TuneIn presets supported')

        if preset.kind == 'speaker':
            preset_type = 1
        else:
            preset_type = 0
        preset_index = int(preset.contentid)

        result = await self.client.request(api_call.set_play_preset(preset_type, preset_index))
        if result.success and self.attribute.muted:
            await self.set_mute(False)

    @is_it_supported
    async def play_url(self, item: UrlMediaItem) -> None:
        """ Play from url.

        Plays a audio stream from an URL. Supported file formats:
        AAC (No DRM), MP3 (max 320kbps), OGG, WMA/WMV (1/2/3/7/9),
        WAV/FLAC/ALAC/AIFF (max 192kHz/24bit).
        NB! The speaker tries to play the body of the request no matter
        what it contains, and does no checking. If the file or body is
        not playable the speaker sometimes freezes, and you will have to
        unplug it to get it to respond again.
        The speaker sends no attributes when an url is played, so it is
        up to the user to set attributes using the UrlMediaItem.

        Arguments:
            item:
                UrlMediaItem with information about the url.
        """
        # Some speakers does't support the SetUrlPlayback API.
        # Maybe it depends on firmware version? We are not checking at
        # the moment.
        # https://sites.google.com/view/samsungwirelessaudiomultiroom/firmware

        item = validate.url_media_item(item)

        # Play url with 2Mb buffer size = 2097152
        result = await self.client.request(api_call.set_url_playback(item.url, 0, 0, 0))

        # *****************************************************
        # Workaround to set attributes not sent by the speaker.
        # *****************************************************
        if result.success:
            self.attribute._title = item.title
            self.attribute._description = item.description
            self.attribute._tracklength = item.duration
            self.attribute._thumbnail = item.thumbnail

        # We need to call all event subscribers after we have manually
        # set new attributes, otherwise the user will not be notified of
        # the change.
        self.events._dispatch_event(True, result)

        if result.success and self.attribute.muted:
            await self.set_mute(False)

    # ******************************************************************
    # Set
    # ******************************************************************
    async def set_mute(self, mute: bool) -> ApiResponse:
        """ Mute or unmute speaker.

        Arguments:
            mute:
                True for mute, False for unmute.

        Returns:
            MuteStatus
        """
        mute = validate.is_boolean(mute)
        return await self.client.request(api_call.set_mute(mute))

    @is_it_supported
    async def set_name(self, name: str) -> ApiResponse:
        """ Set speaker's name.

        Arguments:
            name:
                New speaker name.

        Returns:
            SpkName
        """
        name = validate.name(name)
        return await self.client.request(api_call.set_spk_name(name))

    @is_it_supported
    async def set_repeat_mode(self, repeat_mode: str) -> ApiResponse:
        """ Set repeat mode.

        Arguments:
            repeat_mode:
                One of the following modes: 'all', 'one' or 'off'.

        Returns:
            RepeatMode
        """
        if repeat_mode not in ['all', 'one', 'off']:
            raise ValueError('Repeat mode not valid')
        return await self.client.request(api_call.set_uic_repeat_mode(repeat_mode))

    @is_it_supported
    async def set_shuffle(self, shuffle: bool) -> ApiResponse:
        """ Set shuffle mode.

        Arguments:
            shuffle:
                True for shuffle, False for continues mode.

        Returns:
            ShuffleMode
        """
        shuffle = validate.is_boolean(shuffle)
        return await self.client.request(api_call.set_shuffle_mode(shuffle))

    async def set_volume(self, volume: int) -> ApiResponse:
        """ Set volume level.

        Arguments:
            mute:
                Volume level (0-100).

        Returns:
            VolumeLevel
        """
        volume = validate.volume(volume)
        volume = self.device.encode_volume(volume)
        return await self.client.request(api_call.set_volume(volume))

    # ******************************************************************
    # Select
    # ******************************************************************
    @is_it_supported
    async def select_source(self, source: str) -> None:
        """ Select input source for speaker.

        Arguments:
            source:
                Name of source. One of the following, depending on
                speaker model: 'AUX', 'Bluetooth', 'Coaxial', 'HDMI',
                'HDMI 1', 'HDMI 2', 'Optical', 'TV SoundConnect',
                'USB' or 'Wi-Fi'.
                Available sources: :attr:`Speaker.attribute.source_list`
        """
        source = validate.source(source, self.attribute.source_list)
        source = SOURCES_BY_NAME[source]
        await self.client.request(api_call.set_func(source))

    async def select_sound_mode(self, preset: EqualizerPreset) -> None:
        """ Select sound mode for speaker.

        Arguments:
            preset:
                EqualizerPreset to use. Get available presets with
                :attr:`Speaker.attribute.sound_mode_list`
        """
        preset = validate.equalizer_preset(preset, self.attribute.sound_mode_list)
        await self.client.request(api_call.set_7band_eq_mode(preset.index))

    # ******************************************************************
    # Update
    # ******************************************************************

    async def update(self) -> None:
        """ Update all speaker properties.

        This method should be called after a connection to the speaker
        is established. After that there is normally no need for calling
        this any more, since the state is updated as long as the
        connection to speaker remains.
        """
        # NB! It must be in this order!
        await self.update_speaker_info()
        await self.update_player_info()
        await self.update_media_info()
        await self.update_speaker_settings()

        # Store speaker model specifics
        self.device.update_model(self.attribute._spkmodelname)

    async def update_speaker_info(self) -> None:
        """ Update speaker info. """
        # Main speaker info
        await self.client.request(api_call.get_main_info())
        # Names
        await self.client.request(api_call.get_spk_name())
        if self.attribute._grouptype != 'N':
            await self.client.request(api_call.get_group_name())
        # Firmware version
        await self.client.request(api_call.get_software_version())
        # Network connection
        await self.client.request(api_call.get_ap_info())
        # Connected clients
        # TODO: Get 'IpInfo' - Don't know the API-call to make.

    async def update_player_info(self) -> None:
        """ Update player attributes. """
        await self.client.request(api_call.get_func())
        await self.client.request(api_call.get_volume())
        await self.client.request(api_call.get_mute())
        await self.client.request(api_call.get_shuffle_mode())
        await self.client.request(api_call.get_repeat_mode())
        await self.client.request(api_call.get_current_eq_mode())

    async def update_media_info(self) -> None:
        """ Update current media attributes. """
        if self.attribute._function == 'wifi' and self.attribute._submode == 'cp':
            await self.client.request(api_call.get_radio_info())
        elif self.attribute._function == 'wifi' and self.attribute._submode == 'dlna':
            await self.client.request(api_call.get_music_info())

    async def update_speaker_settings(self) -> None:
        """ Update speaker settings. """
        # Favorites. At the moment only TuneIn presets are supported.
        await self.client.request(api_call.set_select_radio())
        await self.client.request(api_call.get_preset_list(0, 30))
        # Equalizer
        await self.client.request(api_call.get_7band_eq_list())

    # ******************************************************************
    # Get
    # ******************************************************************

    async def get_model(self) -> str:
        """ Get model name of speaker.

        Returns:
            Model name of speaker.

        Raise:
            ApiCallError if call failed.
        """
        response = await self.client.request(api_call.get_main_info())
        if model := response.get_key('spkmodelname'):
            device = get_device_info(model)
            return device.name
        raise ApiCallError

    async def get_mute(self) -> bool:
        """ Get mute state of speaker.

        Returns:
            True if speaker is muted.

        Raise:
            ApiCallError if call failed.
        """
        response = await self.client.request(api_call.get_mute())
        if mute := response.get_key('mute'):
            return (mute == 'on')
        raise ApiCallError

    async def get_name(self) -> str:
        """ Get speaker's name.

        Returns:
            Speaker name.

        Raise:
            ApiCallError if call failed.
        """
        response = await self.client.request(api_call.get_spk_name())
        if name := response.get_key('spkname'):
            return name
        raise ApiCallError

    async def get_source(self) -> str:
        """ Get current selected source on speaker.

        Returns:
            Name of current selected source.

        Raise:
            ApiCallError if call failed.
        """
        response = await self.client.request(api_call.get_func())
        if source := response.get_key('function'):
            return SOURCES_BY_API.get(source, 'Unknown')
        raise ApiCallError

    async def get_speaker_id(self) -> str:
        """ Get speaker ID.

        Returns:
            ID of speaker.

        Raise:
            ApiCallError if call failed.
        """
        response = await self.client.request(api_call.get_device_id())
        if speaker_id := response.get_key('device_id'):
            return speaker_id
        raise ApiCallError

    async def get_tunein_presets(self) -> list[dict[str, str]]:
        """ Get TuneIn presets stored in speaker.

        Returns:
            List of TuneIn presets.

        Raise:
            ApiCallError if call failed.
        """
        await self.client.request(api_call.set_select_radio())
        response = await self.client.request(api_call.get_preset_list(0, 30))
        if response.get_key('cpname') == 'tunein':
            return response.get_subkey('presetlist', 'preset')
        else:
            raise ApiCallError

    async def get_volume(self) -> int:
        """ Get volume level.

        Returns:
            Speaker volume (0-100).

        Raise:
            ApiCallError if call failed.
        """
        response = await self.client.request(api_call.get_volume())
        if volume := response.get_key('volume'):
            return self.device.decode_volume(int(volume))
        raise ApiCallError

    # *********************************************************************************************
    # Group
    # *********************************************************************************************

    async def group(self,
                    slaves_before: list[Speaker],
                    slaves_after: list[Speaker],
                    group_name: str | None = None
                    ) -> None:
        """ Group this speaker.

        Create or edit a speaker groupe. This method can only be called
        when the speaker is already master in the the groupe or if it is
        not in a group.
        If the speaker is already master you must provide all the slaves
        because the speaker is not aware of its slaves.

        Arguments:
            slaves_before:
                List of all the slaves in the group, or empty list if
                the speaker is not yet master.
            slaves_after:
                List of all the slaves in the group, or empty list if
                the group should be deleted.
        """
        # Check that we have needed attributes to call grouping API
        if (
            self.attribute.mac is None or
            self.attribute.name is None or
            self.attribute.is_grouped is None
        ):
            raise PywamError('Speaker attributes not available')

        # If this speaker is slave to another group it can't be master
        # in a new group.
        if self.attribute.is_slave:
            raise PywamError('This speaker is already a slave in another group')

        # Validate all slave speakers
        slaves_before = validate.speakers(slaves_before)
        slaves_after = validate.speakers(slaves_after)

        speakers_to_add = list(set(slaves_after)-set(slaves_before))
        speakers_to_remove = list(set(slaves_before)-set(slaves_after))

        if not self.attribute.is_grouped and len(slaves_before) > 0:
            raise PywamError('Speaker has no slaves now')

        for speaker in slaves_before:
            if speaker.attribute.master_ip != self.ip:
                raise PywamError(f'"{speaker.attribute.name}" is not a slave in this group')

        for speaker in speakers_to_add:
            if speaker.attribute.is_grouped:
                raise PywamError(f'"{speaker.attribute.name}" is already grouped')

        subspeakers = []
        for speaker in slaves_after:
            if speaker.attribute.mac is None:
                raise PywamError(f'"{speaker.attribute.name}" has no MAC')
            else:
                subspeakers.append({'ip': speaker.ip, 'mac': speaker.attribute.mac})

        # Set group name
        if not group_name:
            group_name = f'{self.attribute.name}_group'
        group_name = validate.name(group_name)

        # Remove speakers
        api_calls = []
        for speaker in speakers_to_remove:
            _LOGGER.info('Adding "%s" to be removed', speaker.attribute.name)
            api_calls.append(speaker.client.request(api_call.set_ungroup()))
        _LOGGER.info('Removing %s speakers from group', len(api_calls))
        await asyncio.gather(*api_calls)

        # Delete group if no slaves after
        if len(slaves_after) == 0:
            _LOGGER.info('Deleting group by send ungroup to master')
            await self.client.request(api_call.set_ungroup())
            return

        # Update group
        api = api_call.set_multispk_group_mainspk(
            group_name,
            len(subspeakers) + 1,
            self.attribute.mac,
            self.attribute.name,
            subspeakers,
        )
        _LOGGER.info('Calling master with updated group info')
        _LOGGER.info(' - Subspeakers: %s', subspeakers)
        _LOGGER.info(' - Groupname: %s', group_name)
        await self.client.request(api)

        # Call added slaves
        api_calls = []
        for speaker in slaves_after:
            api = api_call.set_multispk_group_subspk(
                group_name,
                len(subspeakers) + 1,
                self.ip,
                self.attribute.mac,
            )
            _LOGGER.info('%s will soon be called and set to slave', speaker.attribute.name)
            api_calls.append(speaker.client.request(api))
        _LOGGER.info('Calling all slaves with updated information')
        await asyncio.gather(*api_calls)



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Api response handling. """

from collections import defaultdict
from typing import Any
import xml.etree.ElementTree as ET


class ApiResponse:
    """ Messages from the speaker and socket connection.

    Contains speaker events (API responses) and error messages from both
    speaker and methods that are involved in getting responses from the
    speaker.

    Attributes:
        raw_response:
            Raw XML-string from speaker.
        api_type:
            Type of API call. (XML-node: type)
        method:
            Name of response (XML-node: method)
        user:
            Sender of API call. (XML-node: user_identifier)
        version:
            API-version. (XML-node: version)
        speaker_ip:
            Speaker IP address. (XML-node: speakerip)
        success:
            True if call was successful.
            (XML-node "response" attribute "result")
        data:
            Requested data (XML-node: response).
        err_msg:
            Description of what and/or where something went wrong.
        err_repr:
            String representation of the exception that caused the
            error response.
    """
    __slots__ = ('raw_response', 'api_type', 'method', 'user',
                 'version', 'speaker_ip', 'success', 'data',
                 'err_msg', 'err_repr')

    def __init__(self,
                 raw_response: str = '',
                 api_type: str = '',
                 method: str = '',
                 user: str = '',
                 version: str = '',
                 speaker_ip: str = '',
                 success: bool = False,
                 data: Any = None,
                 err_msg: str = '',
                 err_repr: str = '',
                 ) -> None:
        self.raw_response = raw_response
        self.api_type = api_type
        self.method = method
        self.user = user
        self.version = version
        self.speaker_ip = speaker_ip
        self.success = success
        self.data = data
        self.err_msg = err_msg
        self.err_repr = err_repr

    def __str__(self) -> str:
        return ('ApiResponse:\n' +
                f' raw_response: {self.raw_response}\n' +
                f' api_type: {self.api_type}\n' +
                f' method: {self.method}\n' +
                f' user: {self.user}\n' +
                f' version: {self.version}\n' +
                f' speaker_ip: {self.speaker_ip}\n' +
                f' success: {self.success}\n' +
                f' data: {self.data}\n' +
                f' err_msg: {self.err_msg}\n' +
                f' err_repr: {self.err_repr}\n'
                )

    def get_key(self, key: str, default: Any = None) -> Any:
        """ Get a value from data.

        Arguments:
            key:
                Key for value to return.
            default:
                Value to return if key can not be found.

        Returns:
            Value for the given key if the key exists,
            otherwise given default value.
        """
        if isinstance(self.data, dict):
            return self.data.get(key, default)
        return default

    def get_subkey(self, key, subkey, default: Any = None) -> Any:
        """ Get a value from data.

        Arguments:
            key:
                Key in which the sub key should be.
            subkey:
                Key for value to return.
            default:
                Value to return if key can not be found.

        Returns:
            Value for the given key if the key exists,
            otherwise given default value.
        """
        if isinstance(self.data, dict):
            if value := self.data.get(key):
                return value.get(subkey, default)
        return default


def api_decode(raw_response: str) -> ApiResponse:
    """ Parse XML to ApiResponse.

    Arguments:
        raw_response:
            XML response from speaker.

    Returns:
        :class:`ApiResponse` with parsed response from speaker.
    """
    def xml_to_dict(node: ET.Element) -> dict:
        """ Convert an ElementTree.Element to a Python dictionary.

        Arguments:
            node:
                ElementTree.Element to be converted.

        Returns:
            Python dictionary with the converted XML tree.
        """
        resp: dict[str, Any] = {node.tag: {} if node.attrib else None}

        children = list(node)
        if len(node) > 0:
            dd: dict[str, Any] = defaultdict(list)
            for dc in map(xml_to_dict, children):
                for k, v in dc.items():
                    dd[k].append(v)
            resp = {node.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
        if node.attrib:
            resp[node.tag].update(('@' + k, v) for k, v in node.attrib.items())
        if node.text:
            text = node.text.strip() or ''
            if children or node.attrib:
                if text:
                    resp[node.tag]['#text'] = text
            else:
                resp[node.tag] = text
        return resp

    root = ET.fromstring(raw_response)
    root_dict = xml_to_dict(root)
    response_dict = root_dict[root.tag]
    response_dict['type'] = root.tag

    # One response ('GlobalSearch') found where data is in 'async_response'
    # instead of 'response'. Needs more investigation.
    data = response_dict.get('response', {})
    success = (data.get('@result') == 'ok')

    return ApiResponse(raw_response=raw_response,
                       api_type=response_dict.get('type', ''),
                       method=response_dict.get('method', ''),
                       user=response_dict.get('user_identifier', ''),
                       version=response_dict.get('version', ''),
                       speaker_ip=response_dict.get('speakerip', ''),
                       success=success,
                       data=data,
                       )


def api_error(msg: str,
              error: Exception | None = None,
              raw_response: str = '',
              ) -> ApiResponse:
    """ Create ApiResponse from a Exception.

    Arguments:
        msg:
            Description of what and/or where something went wrong.
        error (optional):
            Exception that caused this message to be created.
        raw_response (optional):
            XML response from speaker..

    Returns:
        :class:`ApiResponse` with information about the error that
        occurred.

    """
    return ApiResponse(err_msg=msg,
                       err_repr=repr(error),
                       raw_response=raw_response,
                       )



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Collection of small helper functions for pywam. """


def listify(input_list):
    """ Ensure that the returned object is a list.

    Encapsulates all objects except None and lists in a list.
    """
    if input_list is None:
        return None
    if not input_list:
        return []
    if not isinstance(input_list, list):
        input_list = [input_list]
        return input_list
    else:
        return input_list


def timelength_to_sec(timelength: str) -> int:
    """ Convert API response data timelength to seconds. """

    if not isinstance(timelength, str):
        return 0

    # As for now only response format received is HH:MM:SS.uuu format
    tl = timelength.replace('.', ':')
    try:
        return int(sum(x * int(t) for x, t in
                       zip([0.001, 1, 60, 3600], reversed(tl.split(":"))))
                   )
    except Exception:
        return 0



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Classes and function for handling playing url. """

from __future__ import annotations

from dataclasses import dataclass
from urllib.parse import urlsplit


# https://www.iana.org/assignments/media-types/media-types.xhtml
SUPPORTED_MIME_TYPES = (
    'audio/aac',
    'audio/aacp',
    'audio/aiff',
    'audio/flac',
    'audio/mp4',
    'audio/mp4a-latm',
    'audio/MPA',
    'audio/mpa-robust',
    'audio/mpeg',
    'audio/mpeg4-generic',
    'audio/ogg',
    'audio/vnd.wave',
    'audio/vorbis',
    'audio/vorbis-config',
    'audio/wav',
    'audio/wave',
    'audio/x-aiff',
    'audio/x-flac',
    'audio/x-ms-wma',
    'audio/x-wav',
)


@dataclass
class UrlMediaItem:
    """ Representing a playable url stream.

    Arguments:
        url:
            url with a playable stream.
        title:
            Title of stream.
        description:
            Description for the stream.
        duration (optional):
            Duration in seconds.
        thumbnail (optional):
            Url to thumbnail.
    """

    def __init__(self,
                 url: str,
                 title: str = '',
                 description: str = '',
                 duration: str | None = None,
                 thumbnail: str | None = None
                 ) -> None:
        """ Initialize a UrlMediaItem. """
        self._url = url
        self._title = title
        self._description = description
        self._duration = duration
        self._thumbnail = thumbnail

    @property
    def url(self) -> str:
        """ Return url of media item. """
        return self._url

    @property
    def title(self) -> str:
        """ Return title of media item. """
        if self._title:
            return self._title

        title = 'URL stream'
        try:
            if (host := urlsplit(self._url).hostname) is not None:
                domain = ".".join(host.split('.')[-3:])
                title = domain.replace('www.', '')
        except Exception:
            pass

        return title

    @property
    def description(self) -> str:
        """ Return description of media item. """
        if self._description:
            return self._description

        description = 'URL stream'
        try:
            path = urlsplit(self._url).path
            file = path.split('/')[-1]
            if file:
                description = file
        except Exception:
            pass

        return description

    @property
    def duration(self) -> str | None:
        """ Return duration of media item in seconds. """
        return self._duration

    @property
    def thumbnail(self) -> str | None:
        """ Return thumbnail url of media item. """
        return self._thumbnail



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Samsung WAM API object factory.

All wrapper functions returns a :class:`ApiCall` to be used when calling
:meth:`pywam.client.WamClient.request`.
All known responses to the different API calls are documented in
:module:`pywam.events`.
"""

from __future__ import annotations

import urllib.parse


class ApiCall:
    """ API message to speaker.

    Arguments:
        api_type:
            Type of API call, 'UIC' or 'CPM'.
        method:
            API method name.
        args (optional):
            List of tuples with arguments for the api method.
            First item: Name of the argument.
            Second item: Value of argument.
            Third item: Type hint. ('str'|'dec'|'cdata'|'dec_arr')
        expected_response (optional):
            Expected name of method in response.
        user_check (optional):
            True if response should be checked against user ID of sender.
        timeout_multiple (optional):
            Multiples of API timeout.
    """
    __slots__ = ('api_type', 'method', 'args', 'pwron', 'expected_response',
                 'user_check', 'timeout_multiple')

    # pylint: disable=dangerous-default-value
    def __init__(self,
                 api_type: str,
                 method: str,
                 pwron: bool = False,
                 args: list[tuple[str, str | int | list[int], str]] = [],
                 expected_response: str = '',
                 user_check: bool = False,
                 timeout_multiple: int = 1,
                 ) -> None:
        """ Create a ApiCall object. """
        self.api_type = api_type
        self.method = method
        self.pwron = pwron
        self.args = args
        self.expected_response = expected_response
        self.user_check = user_check
        self.timeout_multiple = timeout_multiple

    def __str__(self):
        return ('ApiCall:\n' +
                f'api_type: {self.api_type}\n' +
                f'method: {self.method}\n' +
                f'pwron: {self.pwron}\n' +
                f'args: {self.args}\n' +
                f'expected_response: {self.expected_response}\n' +
                f'user_check: {self.user_check}\n' +
                f'timeout_multiple: {self.timeout_multiple}\n'
                )

    @property
    def url(self) -> str:
        """ ASCII encoded URL safe string with API call. """
        # pwron
        if self.pwron:
            payload = ['<pwron>on</pwron>']
        else:
            payload = []

        # method
        payload.append(f'<name>{self.method}</name>')

        # arguments
        if self.args:
            for arg in self.args:
                name, value, type_hint = arg
                if type_hint == 'cdata':
                    payload.append(f'<p type="{type_hint}" name="{name}" val="empty">'
                                   f'<![CDATA[{value}]]></p>')
                elif type_hint == 'str_arr' or type_hint == 'dec_arr':
                    if isinstance(value, list):
                        value = ''.join([f'<item>{v}</item>' for v in value])
                    else:
                        value = f'<item>{value}</item>'
                    payload.append(f'<p type="{type_hint}" name="{name}" val="empty">'
                                   f'{value}</p>')
                else:
                    payload.append(f'<p type="{type_hint}" name="{name}" val="{value}"/>')

        # create url
        cmd = ''.join(payload)
        return f'/{self.api_type}?cmd={urllib.parse.quote(cmd)}'


# **********************************************************************
# Speaker info
# **********************************************************************

def get_software_version() -> ApiCall:
    """ (UIC) Get software version on speaker. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetSoftwareVersion',
                   args=[],
                   expected_response='SoftwareVersion',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_spk_name() -> ApiCall:
    """ (UIC) Get speaker name. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetSpkName',
                   args=[],
                   expected_response='SpkName',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_spk_name(spkname: str) -> ApiCall:
    """ (UIC) Set speaker name.

    Arguments:
        spkname:
            New speaker name.
    """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='SetSpkName',
                   args=[('spkname', spkname, 'cdata')],
                   expected_response='SpkName',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_main_info() -> ApiCall:
    """ (UIC) Get main information about speaker.

    Important:
        Api call will cause the following events/responses:
            'RequestDeviceInfo':
                Empty response without any information.
            'MainInfo':
                The response with information about the speaker.
    """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetMainInfo',
                   args=[],
                   expected_response='MainInfo',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_device_id() -> ApiCall:
    """ (CPM) Get speaker ID. """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='GetDeviceId',
                   args=[],
                   expected_response='DeviceId',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_ap_info() -> ApiCall:
    """ (UIC) Get information about network connection. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetApInfo',
                   args=[],
                   expected_response='ApInfo',
                   user_check=False,
                   timeout_multiple=1,
                   )


# **********************************************************************
# Player and media
# **********************************************************************

def get_volume() -> ApiCall:
    """ (UIC) Get current volume level. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetVolume',
                   args=[],
                   expected_response='VolumeLevel',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_volume(volume: int) -> ApiCall:
    """ (UIC) Set speaker volume level.

    Arguments:
        volume:
            Volume level between 0 and 30.
    """
    return ApiCall(api_type='UIC',
                   pwron=True,
                   method='SetVolume',
                   args=[('volume', volume, 'dec')],
                   expected_response='VolumeLevel',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_mute() -> ApiCall:
    """ (UIC) Get mute state of the speaker. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetMute',
                   args=[],
                   expected_response='MuteStatus',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_mute(mute: bool) -> ApiCall:
    """ (UIC) Set mute/unmute state of the speaker.

    Arguments:
         mute:
            True to mute.
    """
    return ApiCall(api_type='UIC',
                   pwron=True,
                   method='SetMute',
                   args=[('mute', 'on' if mute else 'off', 'str')],
                   expected_response='MuteStatus',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_shuffle_mode() -> ApiCall:
    """ (UIC) Retrieve currently set shuffle mode. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetShuffleMode',
                   args=[],
                   expected_response='ShuffleMode',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_shuffle_mode(shuffle_mode: bool) -> ApiCall:
    """ (UIC) Enable/disable shuffle mode of the playlist.

    Arguments:
        shuffle_mode:
            True for shuffle mode
    """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='SetShuffleMode',
                   args=[('shufflemode', 'on' if shuffle_mode else 'off', 'str')],
                   expected_response='ShuffleMode',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_repeat_mode() -> ApiCall:
    """ (UIC) Get playback repeat mode. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetRepeatMode',
                   args=[],
                   expected_response='RepeatMode',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_uic_repeat_mode(repeat_mode: str) -> ApiCall:
    """ (UIC) Set playback repeat mode.

    Arguments:
        mode:
            Repeat mode. ('one' | 'all' | 'off')
    """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='SetRepeatMode',
                   args=[('repeatmode', repeat_mode, 'str')],
                   expected_response='RepeatMode',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_cp_info() -> ApiCall:
    """ (CPM) Get info about currently selected cp service.

    This does not mean that the speaker is playing this cp at the
    moment, only that this is the cp selected for doing operations
    on. E.g: if you want to get TuneIn presets from the speaker but
    the current selected cp is Spotify, you will try to get presets
    from Spotify, which is not supported.
    """
    # TODO: This API can have arguments, but I don't know what and why
    # <p type="str" name="cpname" val="%s"/>
    # <p type="str" name="cpname" val="Spotify"/>
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='GetCpInfo',
                   args=[],
                   expected_response='CpInfo',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_func(func: str) -> ApiCall:
    """ (UIC) Set the source for the speaker.

    Arguments:
        source:
            Source to select.
            ('aux' | 'bt' | 'hdmi' | 'optical' | 'soundshare' | 'wifi')
    """
    return ApiCall(api_type='UIC',
                   pwron=True,
                   method='SetFunc',
                   args=[('function', func, 'str')],
                   expected_response='CurrentFunc',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_func() -> ApiCall:
    """ (UIC) Get current source for the speaker. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetFunc',
                   args=[],
                   expected_response='CurrentFunc',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_uic_playback_control(command: str) -> ApiCall:
    """ (UIC) Pause/resume current playlist.

    This is for UIC services (DLNA, USB...).

    Arguments:
        command:
            'resume' | 'pause'

    Important:
        Api call will cause the following events/responses:
            'PlaybackStatus':
                Confirmation that call is received by speaker.
            'PausePlaybackEvent':
                When playback is paused. Can be received before
                'PlaybackStatus'
            'StartPlaybackEvent':
                When playback is resumed. Can be received before
                'PlaybackStatus'
    """
    return ApiCall(api_type='UIC',
                   pwron=True,
                   method='SetPlaybackControl',
                   args=[('playbackcontrol', command, 'str')],
                   expected_response='PlaybackStatus',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_cpm_playback_control(command: str) -> ApiCall:
    """ (CPM) Play/pause/stop current selected cp.

    This is for CPM apps. (TuneIn, Spotify, Amazon...)
    Not all commands work on all cp.
    TuneIn: play | pause | stop
    Spotify: play | pause

    Arguments:
        command:
            'play' | 'pause' | 'stop'

    Important:
        Api call will cause the following events/responses:
            'PlaybackStatus':
                Confirmation that call is received by speaker.
            'PausePlaybackEvent':
                When playback is paused. Can be received before
                'PlaybackStatus'
            'StartPlaybackEvent':
                When playback is resumed. Can be received before
                'PlaybackStatus'
    """
    return ApiCall(api_type='CPM',
                   pwron=True,
                   method='SetPlaybackControl',
                   args=[('playbackcontrol', command, 'str')],
                   expected_response='PlaybackStatus',
                   user_check=False,
                   timeout_multiple=5,
                   )


def set_trick_mode(trick_mode: str) -> ApiCall:
    """ (UIC) Move to next/previous track on the playlist.

    Arguments:
        trick_mode:
            'next' | 'previous'
    """
    return ApiCall(api_type='UIC',
                   pwron=True,
                   method='SetTrickMode',
                   args=[('trickmode', trick_mode, 'str')],
                   # TODO: Expected response
                   expected_response='',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_radio_info() -> ApiCall:
    """ (CPM) Get information about what is playing.

    This is for CPM apps. (TuneIn, Spotify, Amazon...). Function must be
    'wifi' and submode 'cp'.

    Important:
        For UIC services (DLNA, USB...) you should use
        :func:`get_music_info`
    """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='GetRadioInfo',
                   args=[],
                   expected_response='RadioInfo',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_music_info() -> ApiCall:
    """ (UIC) Get information about what is playing.

    This is for UIC services (DLNA, USB...). If function is 'wifi
    submode must be 'dlna'.

    Important:
        For CPM apps (TuneIn, Spotify, Amazon... ) you should use
        :func:`get_radio_info`
    """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetMusicInfo',
                   args=[],
                   expected_response='MusicInfo',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_url_playback(url: str, buffersize: int, seektime: int, resume: int) -> ApiCall:
    """ (UIC) Play from url.

    Plays a audio file or stream from an URL. Supported file
    formats: AAC (No DRM), MP3 (max 320kbps), OGG,
    WMA/WMV (1/2/3/7/9), WAV/FLAC/ALAC/AIFF (max 192kHz/24bit).
    NB! The speaker tries to play the response body from the request
    no matter what it contains, and does no checking. If the body
    does not contain a playable bytes object the speaker will
    freeze, and you will have to unplug it to get it to respond
    again.

    Arguments:
        url:
            URL to play
        buffersize:
            Buffer size in bytes. strangely playback seems more stable
            when you set buffer to 0. But needs more testing.
        seektime:
            ???
        resume:
            0 = Do not resume previous service
            1 = resume Does not work for TuneIn or Spotify
    """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='SetUrlPlayback',
                   args=[('url', url, 'cdata'),
                         ('buffersize', int(buffersize), 'dec'),
                         ('seektime', int(seektime), 'dec'),
                         ('resume', int(resume), 'dec'), ],
                   expected_response='UrlPlayback',
                   user_check=False,
                   timeout_multiple=5,
                   )


# **********************************************************************
# Favorites
# **********************************************************************

def set_select_radio() -> ApiCall:
    """ (CPM) Select radio.

    Must be selected before sending commands for TuneIn. If you have
    another cp selected get_preset_list will respond with an error.
    Some commands will work, but there will be unwanted side
    effects. For example if you have Spotify selected and playing and
    call to start playing a radio channel, the channel will play, but
    Spotify will not know that you have change service.
    """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='SetSelectRadio',
                   args=[],
                   expected_response='RadioSelected',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_preset_list(start_index: int, list_count: int) -> ApiCall:
    """ (CPM) Get presets stored in speaker.

    Will return an error (- errcode: 73) if not radio is selected.

    Arguments:
        startindex:
            Starting position to retrieve
        listcount:
            Total number of items to retrieve
    """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='GetPresetList',
                   args=[('startindex', start_index, 'dec'),
                         ('listcount', list_count, 'dec'),
                         ],
                   expected_response='PresetList',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_play_preset(preset_type: int,
                    preset_index: int) -> ApiCall:
    """ (CPM) Select preset of a particular index.

    Arguments:
        preset_type:
            1 - speaker, 0 - my
        preset_index:
            Index of get preset list

    Important:
        Api call will cause the following events/responses:
            'RadioInfo':
                Confirmation that call is received by speaker.
            'MediaBufferStartEvent':
                When starting buffering the selected preset.
            'MediaBufferEndEvent':
                When ready buffering.
            'StartPlaybackEvent':
                When actually starting to play selected preset.
    """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='SetPlayPreset',
                   args=[('presettype', int(preset_type), 'dec'),
                         ('presetindex', int(preset_index), 'dec'), ],
                   expected_response='RadioInfo',
                   user_check=False,
                   timeout_multiple=5,
                   )


# **********************************************************************
# Grouping
# **********************************************************************

def get_group_name() -> ApiCall:
    """ (UIC) Get name of speaker group. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetGroupName',
                   args=[],
                   expected_response='GroupName',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_multispk_group_mainspk(name: str,
                               spknum: int,
                               audiosourcemacaddr: str,
                               audiosourcename: str,
                               subspeakers: list[dict[str, str]]
                               ) -> ApiCall:
    """ (UIC) Group speakers.

    This API call should be sent to the master speaker in a group.

    Arguments:
        name:
            Group name.
        spknum:
            Number of speakers in the group.
        audiosourcemacaddr:
            Master speaker MAC address.
        audiosourcename:
            Name of master.
        subspeakers:
            A list of dictionaries containing slave speakers ip and mac address.
            {'ip': ipaddress, 'mac': mac address}

    Notes:
        Following API-attributes does not seem to change and is therefore not
        arguments to this ApiCall factory.

        index:
            "1" always works
        type:
            "main" because this is for the master speaker.
        audiosourcetype:
            Should be "speaker". Have not found any information about
            optional values.
    """
    args = [('name', name, 'cdata'),
            ('index', 1, 'dec'),
            ('type', 'main', 'str'),
            ('spknum', spknum, 'dec'),
            ('audiosourcemacaddr', audiosourcemacaddr, 'str'),
            ('audiosourcename', audiosourcename, 'cdata'),
            ('audiosourcetype', 'speaker', 'str'),
            ]

    for speaker in subspeakers:
        args.extend([('subspkip', speaker['ip'], 'str'), ('subspkmacaddr', speaker['mac'], 'str')])

    return ApiCall(api_type='UIC',
                   pwron=True,
                   method='SetMultispkGroup',
                   args=args,  # type: ignore
                   expected_response='MultispkGroup',
                   user_check=False,
                   timeout_multiple=3,
                   )


def set_multispk_group_subspk(name: str,
                              spknum: int,
                              mainspkip: str,
                              mainspkmacaddr: str,
                              ) -> ApiCall:
    """ (UIC) Group speakers.

    To be sent to all slaves in a group. Seems to be optional!

    Arguments:
        name:
            Group name.
        spknum:
            Number of speakers in the group.
        mainspkip:
            IP address to master speaker.
        mainspkmacaddr:
            MAC  address to master speaker.

    Notes:
        Following API-attributes does not seem to change and is therefore not
        arguments to this ApiCall factory.

        index:
            "1" always works
        type:
            "sub" because this is for the slave speaker.
        audiosourcetype:
            Should be "speaker". Have not found any information about
            optional values.
    """
    return ApiCall(api_type='UIC',
                   pwron=True,
                   method='SetMultispkGroup',
                   args=[('name', name, 'cdata'),
                         ('index', 1, 'dec'),
                         ('type', 'sub', 'str'),
                         ('spknum', spknum, 'dec'),
                         ('mainspkip', mainspkip, 'str'),
                         ('mainspkmacaddr', mainspkmacaddr, 'str'),
                         ],
                   user_check=False,
                   timeout_multiple=3,
                   )


def set_ungroup() -> ApiCall:
    """ (UIC) Ungroup speakers """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='SetUngroup',
                   args=[],
                   expected_response='Ungroup',
                   user_check=False,
                   timeout_multiple=1,
                   )


# **********************************************************************
# Equalizer
# **********************************************************************

def get_current_eq_mode() -> ApiCall:
    """ (UIC) Retrieve current equalizer settings. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='GetCurrentEQMode',
                   args=[],
                   expected_response='CurrentEQMode',
                   user_check=False,
                   timeout_multiple=1,
                   )


def get_7band_eq_list() -> ApiCall:
    """ (UIC) Retrieve stored equalizer presets. """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='Get7BandEQList',
                   args=[],
                   expected_response='7BandEQList',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_7band_eq_mode(presetindex: int) -> ApiCall:
    """ (UIC) Change equalizer values to a saved preset.

    Arguments:
        presetindex:
            Index of preset.
    """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='Set7bandEQMode',
                   args=[('presetindex', presetindex, 'dec')],
                   expected_response='7bandEQMode',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_7band_eq_value(presetindex: int,
                       eqvalue1: int,
                       eqvalue2: int,
                       eqvalue3: int,
                       eqvalue4: int,
                       eqvalue5: int,
                       eqvalue6: int,
                       eqvalue7: int,
                       ) -> ApiCall:
    """ (UIC) Change equalizer values to a new values.

    Arguments:
        presetindex:
            Index of preset to temporarily change.
            Safest is to find out current preset and use that.
        eqvalue1:
            150 Hz, value between -6 and 6.
        eqvalue2:
            300 Hz, value between -6 and 6.
        eqvalue3:
            600 Hz, value between -6 and 6.
        eqvalue4:
            1.2 kHz, value between -6 and 6.
        eqvalue5:
            2.5 kHz, value between -6 and 6.
        eqvalue6:
            5.0 kHz, value between -6 and 6.
        eqvalue7:
            10 kHz Hz, value between -6 and 6.
    """
    return ApiCall(api_type='UIC',
                   pwron=False,
                   method='Set7bandEQValue',
                   args=[('presetindex', presetindex, 'dec'),
                         ('eqvalue1', eqvalue1, 'dec'),
                         ('eqvalue2', eqvalue2, 'dec'),
                         ('eqvalue3', eqvalue3, 'dec'),
                         ('eqvalue4', eqvalue4, 'dec'),
                         ('eqvalue5', eqvalue5, 'dec'),
                         ('eqvalue6', eqvalue6, 'dec'),
                         ('eqvalue7', eqvalue7, 'dec'),
                         ],
                   expected_response='7bandEQValue',
                   user_check=False,
                   timeout_multiple=1,
                   )


# **********************************************************************
# Browsing
# **********************************************************************

def set_locale(locale: str) -> ApiCall:
    """ (CPM) Change speaker locale.

    Only known app to support is TuneIn. Used to get local radio
    stations when browsing.

    Arguments:
        locale:
            (eg: 'en-US', 'de-DE', 'sv-SE' )
    """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='SetLocale',
                   args=[('locale', locale, 'str')],
                   expected_response='LocaleChange',
                   user_check=False,
                   timeout_multiple=1,
                   )


def browse_main(startindex: int, listcount: int) -> ApiCall:
    """ (CPM) Browse TuneIn from the root.

    Arguments:
        startindex:
            Were to start. Should always be 0.
        listcount:
            Number of items to get.
    """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='BrowseMain',
                   args=[('startindex', startindex, 'dec'),
                         ('listcount', listcount, 'dec'),
                         ],
                   expected_response='RadioList',
                   user_check=True,
                   timeout_multiple=1,
                   )


def get_upper_radio_list(startindex: int, listcount: int) -> ApiCall:
    """ (CPM) Browse parent of a browsed TuneIn folder.

    Arguments:
        startindex:
            Were to start. Should always be 0.
        listcount:
            Number of items to get.
    """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='GetUpperRadioList',
                   args=[('startindex', startindex, 'dec'),
                         ('listcount', listcount, 'dec'),
                         ],
                   expected_response='RadioList',
                   user_check=True,
                   timeout_multiple=1,
                   )


def get_current_radio_list(startindex: int, listcount: int) -> ApiCall:
    """ (CPM) Browse current TuneIn folder.

    Arguments:
        startindex:
            Were to start. Should always be 0.
        listcount:
            Number of items to get. Android app use 30 as value.
    """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='GetCurrentRadioList',
                   args=[('startindex', startindex, 'dec'),
                         ('listcount', listcount, 'dec'),
                         ],
                   # TODO: Expected response
                   expected_response='',
                   user_check=True,
                   timeout_multiple=1,
                   )


def get_select_radio_list(contentid: int, startindex: int, listcount: int) -> ApiCall:
    """ (CPM) Select and browse a TuneIn folder.

    Arguments:
        contentid:
            ID of folder to browse. Should be in a folder previous browsed
            with this API call or with :browse_main():.
        startindex:
            Were to start. Should always be 0.
        listcount:
            Number of items to get.
    """
    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='GetSelectRadioList',
                   args=[('contentid', contentid, 'dec'),
                         ('startindex', startindex, 'dec'),
                         ('listcount', listcount, 'dec'),
                         ],
                   # TODO: Expected response
                   expected_response='',
                   user_check=False,
                   timeout_multiple=1,
                   )


def set_play_select(select_item_ids: int | list[int]) -> ApiCall:
    """ (CPM) Plays selected TuneIn or app items.

    Arguments:
        select_item_ids:
            Content id returned by get_upper_radio_list(),
            get_select_radio_list() or get_current_radio_list(),
            or list of content ids
    """
    if isinstance(select_item_ids, list):
        method = 'selectitemids'
        arg_type = 'dec_arr'
    else:
        method = 'selectitemid'
        arg_type = 'dec'

    return ApiCall(api_type='CPM',
                   pwron=False,
                   method='SetPlaySelect',
                   args=[(method, select_item_ids, arg_type)],
                   # TODO: Expected response
                   expected_response='',
                   user_check=False,
                   timeout_multiple=1,
                   )



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Functions for validating user inputs. """
from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING
import uuid

from pywam.lib.equalizer import EqualizerPreset, EqualizerValues
from pywam.lib.media_presets import MediaPreset
from pywam.lib.url import UrlMediaItem

if TYPE_CHECKING:
    from pywam.speaker import Speaker


def speakers(speakers: list[Speaker]) -> list[Speaker]:
    """ Validate a speaker objects.

    Checks that the speaker object has both IP and MAC so that we can use it
    for grouping speakers.

    Arguments:
        speakers:
            List of Speaker objects to validate.
    Returns:
        Validated value.

    Raises:
        TypeError:
            If not an list of Speaker object.
        ValueError:
            If Speaker has no IP or MAC attribute.
    """
    if not isinstance(speakers, list):
        raise TypeError('Speaker must be a list of Speaker object')

    from pywam.speaker import Speaker  # To avoid circular imports
    for speaker in speakers:
        if not isinstance(speaker, Speaker):
            raise TypeError('Speaker must be a Speaker object')
        if not speaker.ip:
            raise ValueError('Speaker has no IP address')
        if not speaker.attribute.mac:
            raise ValueError('Speaker has no MAC address')

    return speakers


def equalizer_preset(preset: EqualizerPreset,
                     preset_list: list[EqualizerPreset],
                     ) -> EqualizerPreset:
    """ Validate equalizer preset.

    Arguments:
        preset:
            EqualizerPreset object to validate.
        preset_list:
            List of presets on the speaker.

    Returns:
        Validated EqualizerPreset.

    Raises:
        TypeError:
            If not an EqualizerPreset object.
        ValueError:
            If preset is not found on the speaker.
    """
    if not preset_list:
        raise ValueError('Can not find any presets on the speaker')
    if not isinstance(preset, EqualizerPreset):
        raise TypeError('Preset must be a EqualizerPreset object')
    for in_list in preset_list:
        if preset.name == in_list.name:
            if preset.index == in_list.index:
                return preset

    raise ValueError('Preset could not be found on the speaker')


def equalizer_values(values: EqualizerValues) -> EqualizerValues:
    """ Validate equalizer values.

    Arguments:
        values:
            EqualizerValues object to validate.

    Returns:
        Validated EqualizerValues.

    Raises:
        TypeError:
            If not an EqualizerValues object or values is not integers.
        ValueError:
            If value is not within range if given.
    """
    if not isinstance(values, EqualizerValues):
        raise TypeError('Preset must be a EqualizerValues object')

    for value in values.values:
        if not isinstance(value, int):
            raise TypeError('Equalizer band value must be a integer')
        if not -6 <= value <= 6:
            raise ValueError('Equalizer band value must be between -6 and 6')

    return values


def is_boolean(val: bool) -> bool:
    """ Is value an boolean.

    Arguments:
        val:
            Value to validate.

    Returns:
        Validated value.

    Raises:
        TypeError:
            If value is not a boolean.
    """
    if not isinstance(val, bool):
        raise TypeError('Value must be a boolean')
    return val


def is_integer(val: int, in_range: tuple[int, int] | None = None) -> int:
    """ Is value an integer.

    Arguments:
        val:
            Value to validate.
        in_range (optional):
            If given the integer should not be less than in_range[0]
            and not greater than in_range[1].

    Returns:
        Validated value.

    Raises:
        TypeError:
            If value is not a integer.
        ValueError:
            If value is not within range if given.
    """
    if not isinstance(val, int):
        raise TypeError('Value must be a integer')
    if in_range:
        if not in_range[0] <= val <= in_range[1]:
            raise ValueError(
                f'Value must be between {in_range[0]} and {in_range[1]}'
            )
    return val


def ip(ip: str) -> str:
    """ Validate IP address.

    Arguments:
        ip:
            IP address to validate.

    Returns:
        IP address.

    Raises:
        TypeError:
            If IP address is not a valid.
        ValueError:
            If IP address is not private.
    """
    if not isinstance(ip, str):
        raise TypeError('Speaker IP must be a string')
    try:
        test_ip = ipaddress.ip_address(ip)
    except Exception:
        raise TypeError('Speaker IP is not a valid IP address')
    if not test_ip.is_private:
        raise ValueError('Speaker IP is not a private IP address')
    return ip


def name(name: str) -> str:
    """ Validate speaker name.

    Arguments:
        name:
            Name.

    Returns:
        Name.

    Raises:
        TypeError:
            If name is not a string.
        ValueError:
            If name is not valid.
    """
    if not isinstance(name, str):
        raise TypeError('Speaker name must be a string')
    if not name.isprintable():
        raise ValueError('Speaker name contains unprintable characters')
    if len(name) > 64:
        raise ValueError('Speaker name to long')
    if len(name) < 1:
        raise ValueError('Speaker name to short')
    return name


def port(port: int) -> int:
    """ Validate TCP port.

    Arguments:
        port: TCP port to validate.

    Returns:
        TCP port.

    Raises:
        TypeError:
            If TCP port is not a integer.
        ValueError:
            IF TCP port is out of range.
    """
    if not isinstance(port, int):
        raise TypeError('Speaker port must be a integer')
    if not 0 < port < 65536:
        raise ValueError('Speaker port is not valid port number')
    return port


def media_preset(preset: MediaPreset, preset_list: list[MediaPreset]) -> MediaPreset:
    """ Validate a media preset.

    Arguments:
        preset:
            Preset to be played.
        preset_list:
            List of presets on the speaker.

    Returns:
        Validated preset.

    Raises:
        TypeError:
            If not an EqualizerPreset object.
        ValueError:
            If preset is not found on the speaker.
    """
    if not preset_list:
        raise ValueError('Can not find any presets on the speaker')
    if not isinstance(preset, MediaPreset):
        raise TypeError('Preset must be a MediaPreset object')
    for in_list in preset_list:
        if preset.title == in_list.title:
            if preset.contentid == in_list.contentid:
                return preset

    raise ValueError('Preset could not be found on the speaker')


def source(source: str, source_list: list[str]) -> str:
    """ Validate source.

    Arguments:
        source:
            Source to validate
        model:
            Speaker model as WAM string
        master:
            Set to True if speaker is master in a group

    Returns:
        Validated source.

    Raises:
        TypeError:
            If source is not a string.
        ValueError:
            If given source is not correct.
    """
    if not isinstance(source, str):
        raise TypeError('Source must be a string')
    if source not in source_list:
        raise ValueError('Speaker does not support this source in this mode')
    return source


def url_media_item(item: UrlMediaItem) -> UrlMediaItem:
    """ Validate url media items.

    Arguments:
        item:
            UrlMediaItem to validate.

    Returns:
        Validated value.

    Raises:
        TypeError:
            If not a UrlMediaItem object.
    """
    if not isinstance(item, UrlMediaItem):
        raise TypeError('Must be a UrlMediaItem object.')
    return item


def user(user: str) -> str:
    """ Validate UUID.

    If a empty string is given as argument, a new UUID will be returned.

    Arguments:
        user:
            UUID to validate or empty string.

    Returns:
        UUID.

    Raises:
        TypeError:
            If UUID is not a string.
        ValueError:
            If UUID is not a valid UUID
    """
    if not isinstance(user, str):
        raise TypeError('User must be a string')
    if not user:
        return str(uuid.uuid1())
    try:
        val_user = uuid.UUID(user)
    except Exception:
        raise ValueError('Given UUID is not valid')
    return str(val_user)


def volume(volume: int) -> int:
    """ Validate speaker volume.

    Arguments:
        volume:
            Volume.

    Returns:
        Volume

    Raises:
        TypeError:
            If volume is not a integer.
        ValueError:
            If volume is out of range.
    """
    if not isinstance(volume, int):
        raise TypeError("Volume must be an integer")
    if not 0 <= volume <= 100:
        raise ValueError("Volume must be between 0 and 100")
    return volume



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Classes and function for handling media presets. """
from __future__ import annotations


class MediaPreset:
    """ Class for storing media presets. """

    def __init__(self,
                 cpname: str,
                 kind: str,
                 title: str,
                 description: str,
                 thumbnail: str,
                 contentid: str,
                 mediaid: str,
                 **kwargs) -> None:
        self._cpname = cpname
        self._kind = kind
        self._title = title
        self._description = description
        self._thumbnail = thumbnail
        self._contentid = contentid
        self._mediaid = mediaid

    @property
    def app(self) -> str:
        """ Return app name. Only 'TuneIn' seems to be supported? """
        return self._cpname

    @property
    def kind(self) -> str:
        """ Return kind. Either 'speaker' or 'my'. """
        return self._kind

    @property
    def title(self) -> str:
        """ Return title of radio station. """
        return self._title

    @property
    def description(self) -> str:
        """ Return station description. """
        return self._description

    @property
    def thumbnail(self) -> str:
        """ Return URL to station logo thumbnail. """
        return self._thumbnail

    @property
    def contentid(self) -> str:
        """ Return number in preset list. Used for playing the preset. """
        return self._contentid

    @property
    def mediaid(self) -> str:
        """ Return ID on TuneIn. """
        return self._mediaid



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Constants.

Project constants that stores information that we can't get from the
speaker.
"""


class Feature():
    """ Supported features. """
    NEXT = 'cmd_next'
    PAUSE = 'cmd_pause'
    PLAY = 'cmd_play'
    PREV = 'cmd_previous'
    STOP = 'cmd_stop'
    PLAY_PRESET = 'play_preset'
    PLAY_BROWSER_ITEM = 'play_browser_item'
    PLAY_URL = 'play_url'
    SET_NAME = 'set_name'
    SET_REPEAT = 'set_repeat_mode'
    SET_SHUFFLE = 'set_shuffle'
    SELECT_SOURCE = 'select_source'


EXC_MESSAGE: dict[str, str] = {
    'cmd_next': 'Next',
    'cmd_pause': 'Pause',
    'cmd_play': 'Play',
    'cmd_previous': 'Previous',
    'cmd_stop': 'Stop',
    'play_preset': 'Playing presets',
    'play_browser_item': 'Playing browser items',
    'play_url': 'Playing a URL',
    'set_name': 'Changing speaker name',
    'set_repeat_mode': 'Repeat mode',
    'set_shuffle': 'Shuffle mode',
    'select_source': 'Selecting source',
}

APP_FEATURES: dict[str, list[str]] = {
    'Pandora': [Feature.STOP, Feature.PLAY],
    'Spotify': [Feature.PLAY],
    'Deezer': [Feature.STOP, Feature.PLAY],
    'Napster': [Feature.STOP, Feature.PLAY],
    '8tracks': [Feature.STOP, Feature.PLAY],
    'iHeartRadio': [Feature.STOP, Feature.PLAY],
    'Rdio': [Feature.STOP, Feature.PLAY],
    'BugsMusic': [Feature.STOP, Feature.PLAY],
    'JUKE': [Feature.STOP, Feature.PLAY],
    '7digital': [Feature.STOP, Feature.PLAY],
    'Murfie': [Feature.STOP, Feature.PLAY],
    'JB HI-FI Now': [Feature.STOP, Feature.PLAY],
    'Rhapsody': [Feature.STOP, Feature.PLAY],
    'Qobuz': [Feature.STOP, Feature.PLAY],
    'Stitcher': [Feature.STOP, Feature.PLAY],
    'MTV Music': [Feature.STOP, Feature.PLAY],
    'Milk Music': [Feature.STOP, Feature.PLAY],
    'Milk Music Radio': [Feature.STOP, Feature.PLAY],
    'MelOn': [Feature.STOP, Feature.PLAY],
    'Tidal HiFi': [Feature.STOP, Feature.PLAY],
    'SiriusXM': [Feature.STOP, Feature.PLAY],
    'Anghami': [Feature.STOP, Feature.PLAY],
    'AmazonPrime': [Feature.STOP, Feature.PLAY],
    'Amazon': [Feature.STOP, Feature.PLAY],
    'TuneIn': [Feature.PLAY],
    'Unknown': [Feature.STOP],
    'DLNA': [Feature.PLAY, Feature.PREV,
             Feature.NEXT, Feature.SET_SHUFFLE, Feature.SET_REPEAT],
    'URL Playback': [],
}

SOURCE_FEATURES: dict[str, list[str]] = {
    'AUX': [],
    'Bluetooth': [],
    'Coaxial': [],
    'HDMI': [],
    'HDMI 1': [],
    'HDMI 2': [],
    'Optical': [],
    'TV SoundConnect': [],
    'USB': [],
    'Wi-Fi': [Feature.PLAY_URL, Feature.PLAY_PRESET],
}

SOURCES_BY_NAME: dict[str, str] = {
    'AUX': 'aux',
    'Bluetooth': 'bt',
    'Coaxial': 'coaxial',
    'HDMI': 'hdmi',
    'HDMI 1': 'hdmi1',
    'HDMI 2': 'hdmi2',
    'Optical': 'optical',
    'TV SoundConnect': 'soundshare',
    'USB': 'usb',
    'Wi-Fi': 'wifi',
}
SOURCES_BY_API = {value: key for key, value in SOURCES_BY_NAME.items()}



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Classes and function for handling playing url. """

from __future__ import annotations
from dataclasses import dataclass
import logging
import re
from urllib.parse import urlsplit

from pywam.lib.url import UrlMediaItem


_LOGGER = logging.getLogger(__name__)


M3U = 'M3U'
XM3U = 'Extended M3U'
PLS = 'PLS'
UNKNOWN = 'Unknown'

# https://www.iana.org/assignments/media-types/media-types.xhtml
PLAYLIST_MIME_TYPES = {
    'application/mpegurl': M3U,
    'application/x-mpegurl': M3U,
    'audio/mpegurl': M3U,
    'audio/x-mpegurl': M3U,
    'application/vnd.apple.mpegurl': M3U,
    'application/vnd.apple.mpegurl.audio': M3U,
    'audio/x-scpls': PLS,
    'text/html': UNKNOWN,
}

PLS_FILE: re.Pattern = re.compile(r'^File(\d+)=(.*)', re.IGNORECASE)
PLS_TITLE: re.Pattern = re.compile(r'^Title(\d+)=(.*)', re.IGNORECASE)
PLS_LENGTH: re.Pattern = re.compile(r'^Length(\d+)=(.*)', re.IGNORECASE)


@dataclass
class PlaylistRow:
    """ Stores rows in playlists. """
    file: str
    title: str
    length: str


def parse_playlist(playlist: str, mime_type: str | None = None) -> list[UrlMediaItem]:
    """ Parse a playlist and return first item as UrlMediaItem.

    Arguments:
        playlist:
            Playlist to parse.
        mime_type (optional):
            Format of the playlist. If left out we will try to guess it.

    Returns:
        Playlist as list of UrlMediaItem.
    """
    if mime_type is None:
        list_format = _find_playlist_format(playlist)
        if list_format == UNKNOWN:
            raise TypeError("Unknown playlist format")
    else:
        list_format = PLAYLIST_MIME_TYPES[mime_type]

    if list_format == M3U or list_format == XM3U:
        _LOGGER.debug("Parsing M3U playlist")
        parsed = m3u_parser(playlist)
    elif list_format == PLS:
        _LOGGER.debug("Parsing PLS playlist")
        parsed = pls_parser(playlist)

    return [UrlMediaItem(url=item.file, title=item.title, duration=item.length) for item in parsed]


def _find_playlist_format(playlist: str) -> str:
    """ Try to find out format of the playlist.

    Arguments:
        playlist:
            Playlist to find format of.

    Returns:
        One of the following constants: M3U , XM3U or PLS
        or UNKNOWN if not a playlist.
    """
    first_line = playlist.split(maxsplit=1)[0]

    # XM3U
    if first_line.startswith('#EXTM3U'):
        return XM3U
    # PLS
    if first_line.startswith('[playlist]'):
        return PLS
    # M3U
    test = urlsplit(first_line)
    # If first line looks like an valid url it is a m3u-playlist
    if test.scheme in ['http', 'https'] and test.netloc != '':
        return M3U

    return UNKNOWN


def pls_parser(pls: str) -> list[PlaylistRow]:
    """ Parse PLS-playlist.

    Supports PLS playlists.
    https://schworak.com/blog/e41/extended-pls-plsv2/
    https://en.wikipedia.org/wiki/PLS_(file_format)

    Arguments:
        pls:
            pls-formatted playlist as string.

    Returns:
        A list of dictionaries with the parsed playlist.
    """
    # Streamline line brakes and remove empty lines.
    fixed_list = pls.replace('\r\n', '\n')
    pls_lines = [line for line in fixed_list.split('\n') if line != '']

    # Header should always be [playlist].
    if pls_lines[0].lower() == '[playlist]':
        pls_lines.pop(0)
    else:
        raise TypeError('Not a valid PLS playlist')

    item_number = 1
    file, title, length = ('', '', '')
    number_of_entries, version = (None, None)
    playlist = []

    for line in pls_lines:

        rx_file = PLS_FILE.search(line)
        # First file row in a block
        if rx_file and not file:
            # If the number doesn't match leave loop
            if not int(rx_file.group(1)) == item_number:
                raise TypeError('Playlist items not in correct order')
            file = rx_file.group(2).strip()
            continue
        # New file row
        if rx_file and file:
            # Store previous block
            playlist.append(PlaylistRow(file, title, length))
            # Start new block
            file = rx_file.group(2).strip()
            title, length = ('', '')
            item_number += 1
            continue
        # A block should always start with File statement
        if not file:
            continue

        rx_title = PLS_TITLE.search(line)
        if rx_title:
            # If the number doesn't match leave loop
            if not int(rx_title.group(1)) == item_number:
                raise TypeError('Playlist items not in correct order')
            title = rx_title.group(2).strip()
            continue

        rx_length = PLS_TITLE.search(line)
        if rx_length:
            # If the number doesn't match leave loop
            if not int(rx_length.group(1)) == item_number:
                raise TypeError('Playlist items not in correct order')
            length = rx_length.group(2).strip()
            continue

        # Valid footers are 'NumberOfEntries' and 'Version'
        if line.lower().startswith('NumberOfEntries'):
            if number_of_entries:
                raise TypeError('NumberOfEntries occurs more than once')
            number_of_entries = int(line.rsplit('=')[0])
        if line.lower() == 'Version':
            if version:
                raise TypeError('Version occurs mote than once')
            version = int(line.rsplit('=')[0])

        # File should end after footers
        if number_of_entries and version:
            break

    # Store last iteration
    if file:
        playlist.append(PlaylistRow(file, title, length))
        item_number += 1

    # Check if correct number of entries.
    if number_of_entries:
        if number_of_entries != (item_number - 1):
            raise TypeError('Number of entries mismatch')

    return playlist


def m3u_parser(m3u: str) -> list[PlaylistRow]:
    """ Parse m3u playlist.

    Supports legacy m3u,extended m3u and m3u8.
    https://tools.ietf.org/html/rfc8216
    https://en.wikipedia.org/wiki/M3U

    Arguments:
        m3u:
            m3u-formatted playlist as string.

    Returns:
        List of items in the playlist
    """

    # Lines in a Playlist file are terminated by either a single line feed
    # character (\n) or a carriage return character followed by a line feed
    # character (\r\n). Each line is a URI, is blank, or starts with the
    # character '#'.  Blank lines are ignored.  Whitespace MUST NOT be
    # present, except for elements in which it is explicitly specified.
    fixed_list = m3u.replace('\r\n', '\n')
    m3u_lines = [line for line in fixed_list.split('\n') if line != '']

    # The EXTM3U tag indicates that the file is an Extended M3U [M3U]
    # Playlist file.  It MUST be the first line of every Media Playlist and
    # every Master Playlist. Its format is: '#EXTM3U'
    if m3u_lines[0].upper() == '#EXTM3U':
        # This is a Extended m3u
        m3u_lines.pop(0)

    title, duration = ('', '')
    playlist = []

    for line in m3u_lines:
        # The EXTINF tag specifies the duration of a Media Segment.  It applies
        # only to the next Media Segment.  This tag is REQUIRED for each Media
        # Segment.  Its format is: #EXTINF:<duration>,[<title>] where duration
        # is a decimal-floating-point or decimal-integer number that specifies
        # the duration of the Media Segment in seconds. The remainder of the line
        # following the comma is an optional human readable informative title of
        # the Media Segment expressed as UTF-8 text.
        if line.startswith('#EXTINF:'):
            extinf = line[8:].split(',', 1)
            duration = extinf[0]
            title = extinf[1]
            continue

        # Lines that start with the character '#' are either comments or tags.
        # Tags begin with #EXT. They are case sensitive. All other lines that
        # begin with '#' are comments and SHOULD be ignored.
        if line.startswith('#'):
            # Skip all comments
            continue
        else:
            # Only add items with plausible url to the playlist.
            test = urlsplit(line)
            if test.scheme in ['http', 'https'] and test.netloc != '':
                playlist.append(PlaylistRow(line, title, duration))
                title, duration = ('', '')

    return playlist



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Classes and function for handling the equalizer. """
from __future__ import annotations


class EqualizerPreset:
    """ Stored equalizer presets on the speaker. """

    def __init__(self, presetindex: str, presetname: str, **kwargs) -> None:
        self._presetindex = presetindex
        self._presetname = presetname

    @property
    def name(self) -> str:
        """ Return name of preset. """
        return self._presetname

    @property
    def index(self) -> int:
        """ Return index of preset. """
        return int(self._presetindex)


class EqualizerValues:
    """ Equalizer values.

    Attributes:
        hz_150:
            150 Hz, value between -6 and 6.
        hz_300:
            300 Hz, value between -6 and 6.
        hz_600:
            600 Hz, value between -6 and 6.
        hz_1200:
            1.2 kHz, value between -6 and 6.
        hz_2500:
            2.5 kHz, value between -6 and 6.
        hz_5000:
            5.0 kHz, value between -6 and 6.
        hz_10000:
            10 kHz, value between -6 and 6.
    """

    def __init__(self,
                 hz_150: str,
                 hz_300: str,
                 hz_600: str,
                 hz_1200: str,
                 hz_2500: str,
                 hz_5000: str,
                 hz_10000: str) -> None:
        self._hz_150 = hz_150
        self._hz_300 = hz_300
        self._hz_600 = hz_600
        self._hz_1200 = hz_1200
        self._hz_2500 = hz_2500
        self._hz_5000 = hz_5000
        self._hz_10000 = hz_10000

    @property
    def hz_150(self) -> int:
        """" Return 150 Hz eq setting, value between -6 and 6."""
        return int(self._hz_150)

    @property
    def hz_300(self) -> int:
        """" Return 300 Hz eq setting, value between -6 and 6."""
        return int(self._hz_300)

    @property
    def hz_600(self) -> int:
        """" Return 600 Hz eq setting, value between -6 and 6."""
        return int(self._hz_600)

    @property
    def hz_1200(self) -> int:
        """" Return 1.2 kHz eq setting, value between -6 and 6."""
        return int(self._hz_1200)

    @property
    def hz_2500(self) -> int:
        """" Return 2.5 kHz eq setting, value between -6 and 6."""
        return int(self._hz_2500)

    @property
    def hz_5000(self) -> int:
        """" Return 5.0 kHz eq setting, value between -6 and 6."""
        return int(self._hz_5000)

    @property
    def hz_10000(self) -> int:
        """" Return 10 kHz eq setting, value between -6 and 6."""
        return int(self._hz_10000)

    @property
    def values(self) -> list[int]:
        """ Return all values in a list. """
        return [self.hz_150,
                self.hz_300,
                self.hz_600,
                self.hz_1200,
                self.hz_2500,
                self.hz_5000,
                self.hz_10000,
                ]



# This file is part of the pywam project.
# Copyright (c) Daniel Jönsson. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in project root.

""" Websocket connection to the speaker.

The speaker normally listens on port 55001 for http requests containing
the api call. It is a simple http get request, where the path is the
complete request and header contains user information. The response of
the request is a http message where the body contains the information
in xml-format.
As long as you are connected to the speaker on this port, all requests
from other users of the speaker will be received, as well as all changes
of the speakers state.
"""
from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Callable, Set

from pywam.lib.api_response import ApiResponse, api_decode, api_error
from pywam.lib.exceptions import ApiCallTimeoutError
from pywam.lib.http import build_http_header, parse_stream
from pywam.lib.validate import is_integer


if TYPE_CHECKING:
    from pywam.lib.api_call import ApiCall
    from pywam.lib.http import HttpResponse
    from pywam.speaker import Speaker

_LOGGER = logging.getLogger(__name__)


class WamClient:
    """ Websocket connection to speaker.

    Arguments:
        ip:
            Speaker's IP address. (e.g "192.168.1.100")
        port:
            TCP port that speaker listens on.
        user:
            UUID for the connected user.
    """

    def __init__(self, speaker: Speaker) -> None:
        self._ip = speaker.ip
        self._port = speaker.port
        self._user = speaker.user
        # Socket
        self._connection_timeout: int = 5
        self._event_reader: asyncio.StreamReader | None = None
        self._event_writer: asyncio.StreamWriter | None = None
        self._connecting: bool = False
        self._disconnecting: bool = False
        # Events listener
        self._listening: asyncio.Event = asyncio.Event()
        self._listener_task: asyncio.Task | None = None
        # Event and response handling
        self._subscribers: Set[Callable] = set()
        self._response_queue: asyncio.Queue[ApiResponse] | None = None
        # API requests
        self._http_request: str = build_http_header(self._ip, self._port, self._user)
        self._request_timeout: int = 10

    @property
    def is_connected(self) -> bool:
        """ Returns true when connected to the speaker. """
        if self._event_reader and self._event_writer:
            return True
        return False

    @property
    def is_listening(self) -> bool:
        """ Returns true when listening for responses from speaker. """
        return self._listening.is_set()

    @property
    def connection_timeout(self) -> int:
        """ Time out for connecting to speaker. """
        return self._connection_timeout

    @connection_timeout.setter
    def connection_timeout(self, timeout: int) -> None:
        is_integer(timeout, (5, 60))
        self._connection_timeout = timeout

    @property
    def request_timeout(self) -> int:
        """ Time out for sending API calls to speaker. """
        return self._request_timeout

    @request_timeout.setter
    def request_timeout(self, timeout: int) -> None:
        is_integer(timeout, (5, 60))
        self._request_timeout = timeout

    async def connect(self) -> None:
        """ Connect to speaker.

        Raises:
            ConnectionError: If connection is not established.
        """
        if self._connecting:
            raise ConnectionError('Already trying to connect to speaker')
        if self.is_connected:
            raise ConnectionError('Already connected to speaker')
        if self._disconnecting:
            raise ConnectionError('Waiting for speaker to disconnect')
        self._connecting = True

        try:
            self._event_reader, self._event_writer = await asyncio.wait_for(
                asyncio.open_connection(self._ip, self._port),
                timeout=self.connection_timeout,
            )
        except asyncio.TimeoutError as exc:
            await self.disconnect()
            raise ConnectionError('Timeout when trying to connect to speaker') from exc
        except Exception as exc:
            await self.disconnect()
            raise ConnectionError('Could not connect to speaker') from exc
        finally:
            self._connecting = False

    async def start_listening(self) -> None:
        """ Start listen to events from the speaker. """
        if not self.is_connected:
            raise ConnectionError('Not connected to the speaker')
        if self._listener_task:
            raise ConnectionError('Listener is already running')

        self._listener_task = asyncio.create_task(self._receive_loop())
        await self._listening.wait()

    async def disconnect(self):
        """ Disconnect from speaker. """
        if self._disconnecting:
            return
        self._disconnecting = True

        if self._event_writer:
            try:
                self._event_writer.close()
                await self._event_writer.wait_closed()
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception('(%s) Unhandled exception when disconnecting', self._ip)

        self._event_reader = None
        self._event_writer = None

        self._disconnecting = False

    async def stop_listening(self):
        """ Stop listen to events from the speaker. """
        if self._listener_task:
            try:
                self._listener_task.cancel()
                await self._listener_task
            except asyncio.CancelledError:
                pass
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception('(%s) Unhandled exception when canceling listener task',
                                  self._ip)
            finally:
                self._listener_task = None
        self._listening.clear()

    async def request(self, api_call: ApiCall) -> ApiResponse:
        """ Send API to the speaker.

        Arguments:
            api_call:
                ApiCall object to send to the speaker.

        Returns:
            ApiResponse object with the requested information.

        Raises:
            ConnectionError: If not connected to the speaker.
            ApiCallTimeoutError: If correct response was not received.
        """
        if not self.is_connected or not self.is_listening:
            raise ConnectionError('No connection with speaker')

        # Make sure that a second call is not made until we receive a
        # correct answer or get time out on current call.
        while self._response_queue:
            await asyncio.sleep(1)

        # Start a writer
        # We need to use new StreamWriter at every call to not get EOF
        # in the StreamReader used for listening for events. This could
        # probably be solved by writing a custom protocol in the future,
        # but is not prioritized at the moment.
        try:
            _, writer = await asyncio.wait_for(
                asyncio.open_connection(self._ip, self._port),
                timeout=self.connection_timeout,
            )
        except Exception as e:
            await self.disconnect()
            raise ConnectionError('Could not connect to speaker') from e

        # Send to the speaker.
        request = f'GET {api_call.url}{self._http_request}'
        writer.write(request.encode())
        await writer.drain()

        # For some API the speaker has no response.
        if not api_call.expected_response:
            return ApiResponse()

        # Wait for correct response.
        self._response_queue = asyncio.Queue(1)
        try:
            response = await asyncio.wait_for(
                self._wait_for_response(api_call),
                timeout=self._request_timeout * api_call.timeout_multiple)
        except asyncio.TimeoutError:
            raise ApiCallTimeoutError('No response from speaker')
        except Exception as e:
            raise ApiCallTimeoutError('Error while waiting for response') from e
        finally:
            # Reset writer and queue
            writer.close()
            await writer.wait_closed()
            self._response_queue = None

        return response

    async def _receive_loop_error(self) -> None:
        """ Cancel the receive loop task and disconnect. """
        await self.stop_listening()
        await self.disconnect()

    async def _receive_loop(self) -> None:
        """ Listen for api responses from the speaker.

        This method should be run as a asyncio task for receiving
        responses (events and api requests) from the speaker.
        """
        buffer: bytes = b''
        self._listening.set()
        try:
            while True:
                # If there is a problem with the StreamReader we need
                # to disconnect, so that the user can check this.
                if not self._event_reader:
                    break
                data = await self._event_reader.read(1024)
                # If we receive an empty byte something is wrong!
                if not data:
                    _LOGGER.warning('(%s) Disconnecting because listener received EOF', self._ip)
                    break
                _LOGGER.debug('(%s) Listener received data:', self._ip)
                _LOGGER.debug(data)
                data = buffer + data
                http_responses = parse_stream(data)
                for http_response in http_responses:
                    if http_response.body:
                        await self._response_handler(http_response)
                    buffer = http_response.remainder

        except asyncio.CancelledError:
            _LOGGER.debug('(%s) Receive loop is cancelled', self._ip)
            raise
        except Exception:  # pylint: disable=broad-except
            _LOGGER.exception('(%s) Disconnecting due to error while reading stream', self._ip)

        await self._receive_loop_error()

    async def _response_handler(self, http_response: 'HttpResponse') -> None:
        """ Handles incoming HTTP responses.

        Incoming http responses (XML) get parsed to ApiResponse objects.
        The ApiResponse object is then dispatched to any subscriber,
        and returned to the request method.

        Arguments:
            http_response:
                HttpResponse object containing an response from the
                speaker.
        """
        try:
            api_response = api_decode(http_response.body)
        except Exception as e:  # pylint: disable=broad-except
            api_response = api_error('ApiDecodingError', e, http_response.body)
            _LOGGER.warning('(%s) Error when decoding api response', self._ip)
        self._dispatch_event(api_response)
        await self._return_response(api_response)

    async def _return_response(self, api_response: 'ApiResponse') -> None:
        """ Return the ApiResponse.

        If the request method is awaiting a response the ApiResponse
        object is put in a queue so that it can be returned to the
        caller.

        Argument:
            api_response:
                ApiResponse object to be returned to caller.
        """
        if self._response_queue:
            await self._response_queue.put(api_response)

    async def _wait_for_response(self, api_call: 'ApiCall') -> 'ApiResponse':
        """ Waits for the correct response from speaker.

        Argument:
            api_call:
                ApiCall object that was sent to the speaker

        Returns:
            ApiResponse that was requested.

        Raises:
            RuntimeError: If there is no asyncio queue to put the
                response in.
        """
        if not self._response_queue:
            raise RuntimeError('asyncio.Queue not initialized')
        while True:
            api_response = await self._response_queue.get()
            if api_response.method != api_call.expected_response:
                continue
            if not api_call.user_check:
                return api_response
            if api_response.user == self._user:
                return api_response

    # ******************************************************************
    # Observation methods
    # ******************************************************************

    def _dispatch_event(self, api_response: 'ApiResponse') -> None:
        """ Send events to subscriber.

        Arguments:
            api_response:
                Response from speaker to send to subscriber.
        """
        for subscriber in self._subscribers:
            try:
                subscriber(api_response)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception('(%s) Error when trying to dispatch an event', self._ip)

    def register_subscriber(self, callback: Callable[['ApiResponse'], Any]) -> None:
        """ Register a function for receiving events from speaker.

        Arguments:
            callback:
                Callable that takes one argument to be called when an
                event is received. The event is passed as the first
                argument.

        Raises:
            TypeError: If given argument is not a callable.
        """
        if not callable(callback):
            raise TypeError('Subscriber must be a callable')
        self._subscribers.add(callback)

    def unregister_subscriber(self, callback: Callable) -> None:
        """ Unregister listener.

        Arguments:
            callback:
                Callable to be removed from list of event subscribers.
        """
        self._subscribers.discard(callback)
